from __future__ import annotations
from abc import abstractmethod
import sys
from pydantic import BaseModel
from mongo_tooling_metrics import register_hook

from mongo_tooling_metrics.errors import InvalidMetricsSetup
from mongo_tooling_metrics.lib.hooks import ExitHook
from mongo_tooling_metrics.lib.utils import should_collect_internal_metrics


class TopLevelMetrics(BaseModel):
    """Base class for a top-level metrics objects."""

    @classmethod
    @abstractmethod
    def generate_metrics(cls, *args, **kwargs) -> TopLevelMetrics:
        """Class method to generate this metrics object -- will be executed before process exit."""
        raise InvalidMetricsSetup(
            "'generate_metrics' will be used to construct the top-level metrics object and must be defined."
        )

    @staticmethod
    def should_collect_metrics() -> bool:
        """Determine whether metrics collection should even be registered or not."""
        # Default to collecting metrics on internal MongoDB virtual workstations only -- this can be overwritten.
        return should_collect_internal_metrics()

    @staticmethod
    def initialize_hooks() -> None:
        """Initialize any hooks that these metrics rely on here -- this will get called after the metrics are registered."""
        # Default to registering the ExitHook -- this can be overwritten.
        sys.exit = register_hook(ExitHook(original_fn=sys.exit))

    def is_malformed(self) -> bool:
        """Determine whether these metrics are malformed (have all expected fields/data)."""
        # Default to metrics not being malformed -- this can be overwritten.
        return False


class SubMetrics(BaseModel):
    """Base class for sub-level metrics objects."""

    @classmethod
    @abstractmethod
    def generate_metrics(cls, *args, **kwargs) -> SubMetrics:
        """Class method to generate this metrics object -- will be executed before process exit."""
        raise InvalidMetricsSetup(
            "'generate_metrics' should be used to construct the sub-metrics object and must be defined."
        )

    def is_malformed(self) -> bool:
        """Determine whether these metrics are malformed (have all expected fields/data)."""
        return False



"""Common sub metrics that are used internally at MongoDB."""

import argparse
import configparser
from datetime import datetime
import multiprocessing
import optparse
import os
import socket
import sys
from typing import Any, Dict, List, Optional, Union

import distro
import git
from mongo_tooling_metrics.base_metrics import SubMetrics


class HostInfo(SubMetrics):
    """Class to store host information."""

    ip_address: Optional[str]
    host_os: str
    num_cores: int
    memory: Optional[float]

    @classmethod
    def generate_metrics(cls):  # pylint: disable=arguments-differ
        """Get the host info to the best of our ability."""
        try:
            ip_address = socket.gethostbyname(socket.gethostname())
        except:
            ip_address = None
        try:
            memory = cls._get_memory()
        except:
            memory = None
        return cls(
            ip_address=ip_address,
            host_os=distro.name(pretty=True),
            num_cores=multiprocessing.cpu_count(),
            memory=memory,
        )

    @staticmethod
    def _get_memory():
        """Get total memory of the host system."""
        return os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES') / (1024.**3)

    def is_malformed(self) -> bool:
        """Confirm whether this instance has all expected fields."""
        return None in [self.memory, self.ip_address]


MODULES_FILEPATH = 'src/mongo/db/modules'
CURRENT_DIR = '.'


class GitInfo(SubMetrics):
    """Class to store git repo information."""

    filepath: str
    commit_hash: Optional[str]
    branch_name: Optional[str]
    repo_name: Optional[str]

    @classmethod
    def generate_metrics(cls, filepath: str = CURRENT_DIR):
        """Get the git info for a repo to the best of our ability."""
        try:
            commit_hash = git.Repo(filepath).head.commit.hexsha
        except:
            commit_hash = None
        try:
            if git.Repo(filepath).head.is_detached:
                branch_name = commit_hash
            else:
                branch_name = git.Repo(filepath).active_branch.name
        except:
            branch_name = None
        try:
            repo_name = git.Repo(filepath).working_tree_dir.split("/")[-1]
        except:
            repo_name = None
        return cls(
            filepath=filepath,
            commit_hash=commit_hash,
            branch_name=branch_name,
            repo_name=repo_name,
        )

    @classmethod
    def modules_generate_metrics(cls, modules_filepath: str = MODULES_FILEPATH):
        """Get git info for all modules in the path."""
        modules_git_info = []
        try:
            modules_git_info = [
                cls.generate_metrics(os.path.join(modules_filepath, module))
                for module in os.listdir(modules_filepath)
                if os.path.isdir(os.path.join(modules_filepath, module))
            ]
        except:
            pass
        return modules_git_info

    def is_malformed(self) -> bool:
        """Confirm whether this instance has all expected fields."""
        return None in [self.commit_hash, self.branch_name, self.repo_name]


class BuildInfo(SubMetrics):
    """Class to store build artifact data."""

    build_artifacts: Optional[List[str]]
    artifact_dir: Optional[str]

    @classmethod
    def generate_metrics(
        cls,
        utc_starttime: datetime,
        artifact_dir: str,
    ):
        """Get build artifact data to the best of our ability."""
        return cls(
            build_artifacts=cls.get_artifacts(utc_starttime, artifact_dir),
            artifact_dir=artifact_dir,
        )

    @staticmethod
    def get_artifacts(utc_starttime: datetime, artifact_dir: str) -> List[str]:
        """Search a directory recursively for all artifacts created after the given timestamp."""
        try:
            start_timestamp = datetime.timestamp(utc_starttime)
            artifacts = []
            for root, _, files in os.walk(artifact_dir):
                for file in files:
                    filepath = os.path.join(root, file)
                    _, ext = os.path.splitext(filepath)
                    if ext in ['.a', '.so', ''] and os.path.getmtime(filepath) >= start_timestamp:
                        artifacts.append(filepath)
            return artifacts
        except:
            return None

    def is_malformed(self) -> bool:
        """Confirm whether this instance has all expected fields."""
        return None in [self.artifact_dir, self.build_artifacts]


class CommandInfo(SubMetrics):
    """Class to store command-line information."""

    command: List[str]
    options: Optional[Dict[str, Any]]
    positional_args: Optional[List[str]]

    @classmethod
    def generate_metrics(cls, parser: Union[argparse.ArgumentParser, optparse.OptionParser]):
        """Get command line information to the best of our ability."""
        try:
            # The SCons parser is of type 'optparse.OptionParser' which does not
            # have 'parse_known_args' but has 'parse_args' -- which has the same effect
            known_args, unknown_args = parser.parse_known_args(sys.argv[1:]) if isinstance(
                parser, argparse.ArgumentParser) else parser.parse_args(sys.argv[1:])
            return cls(command=sys.argv, options=vars(known_args), positional_args=unknown_args)
        except:
            return cls(command=sys.argv, options=None, positional_args=None)

    def is_malformed(self) -> bool:
        """Confirm whether this instance has all expected fields."""
        return None in [self.options, self.positional_args]


SCONS_ENV_FILE = "scons_env.env"
SCONS_SECTION_HEADER = "SCONS_ENV"


class SConsInfo(SubMetrics):
    """Class to store the SCons specific information."""

    env: Optional[Dict[str, Any]]

    @classmethod
    def generate_metrics(
        cls,
        artifact_dir: str,
        env_vars: "SCons.Variables.Variables",
        env: "SCons.Script.SConscript.SConsEnvironment",
    ):
        """Get SCons info to the best of our ability."""
        return cls(env=cls._get_scons_env_vars_dict(artifact_dir, env_vars, env))

    @staticmethod
    def _get_scons_env_vars_dict(
        artifact_dir: str,
        env_vars: "SCons.Variables.Variables",
        env: "SCons.Script.SConscript.SConsEnvironment",
    ) -> Optional[Dict[str, Any]]:
        """Get the environment variables options that can be set by users."""

        scons_env_filepath = os.path.join(artifact_dir, SCONS_ENV_FILE)
        try:
            # Use SCons built-in method to save environment variables to a file
            env_vars.Save(scons_env_filepath, env)

            # Add a section header to the file so we can easily parse with ConfigParser
            with open(scons_env_filepath, 'r') as original:
                data = original.read()
            with open(scons_env_filepath, 'w') as modified:
                modified.write(f"[{SCONS_SECTION_HEADER}]\n" + data)

            # Parse file using config parser
            config = configparser.ConfigParser()
            config.read(scons_env_filepath)
            str_dict = dict(config[SCONS_SECTION_HEADER])
            return {key: eval(val) for key, val in str_dict.items()}  # pylint: disable=eval-used
        except:
            return None

    def is_malformed(self) -> bool:
        """Confirm whether this instance has all expected fields."""
        return self.env is None



import atexit
from enum import Enum
import logging
from typing import Any, Dict, Type

from pydantic import BaseModel
import pydantic
import pymongo
from mongo_tooling_metrics.base_metrics import TopLevelMetrics
from mongo_tooling_metrics.lib.utils import should_collect_internal_metrics
from mongo_tooling_metrics.errors import ExternalHostException, InvalidMetricsObject, InvalidMetricsSetup, MetricsCollectionFailure

logger = logging.getLogger('tooling_metrics')
METRICS_FAILURE_LOG = "\nMetrics Collection Failed -- execution of this program should not be affected.\n"


class MetricsVerbosity(Enum):
    """Enum for determining verbosity of metrics collection."""
    RAISE = "raise"
    LOG = "log"
    SILENT = "silent"


class MongoMetricsClient(BaseModel):
    """Client used to insert metrics to the target collection."""

    # TODO: Find a better way to type mongo_client
    mongo_client: Any
    db_name: str
    collection_name: str
    verbosity: MetricsVerbosity

    def register_metrics(
        self,
        metrics_class: Type[TopLevelMetrics],
        **kwargs,
    ) -> None:
        """Register the metrics to be generated and persisted at process exit -- kwargs will be passed to 'generate_metrics'."""
        if not issubclass(metrics_class, TopLevelMetrics):
            raise InvalidMetricsSetup("Metrics classes must inherit from 'TopLevelMetrics'.")
        if not metrics_class.should_collect_metrics():
            return
        metrics_class.initialize_hooks()
        atexit.register(
            self._verbosity_enforced_save_metrics,
            metrics_class,
            **kwargs,
        )

    def _verbosity_enforced_save_metrics(self, metrics_class: Type[TopLevelMetrics],
                                         **kwargs) -> None:
        """Save metrics and log/raise according to verbosity."""
        try:
            metrics_dict = self._get_metrics_as_dict(metrics_class, **kwargs)
            self._save_metrics(metrics_dict)
        except Exception as _:
            if self.verbosity == MetricsVerbosity.LOG:
                logger.warning(METRICS_FAILURE_LOG)
            elif self.verbosity == MetricsVerbosity.RAISE:
                logger.exception(METRICS_FAILURE_LOG)

    def _get_metrics_as_dict(self, metrics_class: Type[TopLevelMetrics],
                             **kwargs) -> Dict[str, Any]:
        """Create metrics object & convert to dictionary."""
        try:
            return metrics_class.generate_metrics(**kwargs).dict()
        except pydantic.error_wrappers.ValidationError as exc:
            raise InvalidMetricsObject(
                f"Could not generate metrics due to Pydantic validation error. The actual data types do not match the expected data types."
            ) from exc
        except Exception as exc:
            raise InvalidMetricsObject(
                f"Could not create user defined metrics object. You can use the following command to validate a metrics object: '{metrics_class.__name__}.generate_metrics(**kwargs).dict()'."
            ) from exc

    def _save_metrics(self, metrics_dict: Dict[str, Any]) -> None:
        """Save metrics to the atlas cluster."""
        try:
            with pymongo.timeout(1):
                self.mongo_client[self.db_name][self.collection_name].insert_one(metrics_dict)
        except pymongo.errors.PyMongoError as exc:
            if exc.timeout:
                raise MetricsCollectionFailure(
                    "Database insert operation took longer than 1 second.") from exc
            else:
                raise MetricsCollectionFailure(
                    "Database insert operation failed with pymongo error.") from exc
        except TypeError as exc:
            raise MetricsCollectionFailure(
                f"Could not save metrics to mongo cluster. Some attribute type(s) may not be supported: '{metrics_dict}'"
            ) from exc
        except Exception as exc:  # pylint: disable=broad-except
            raise MetricsCollectionFailure(f"Unexpected Metrics Collection Failure.") from exc


# This is a MongoDB internal cluster used to collect internal tooling metrics and usage.
# This cluster is only accessible from MongoDB employee virtual workstations.
# Tooling metrics are only collected for MongoDB employees.

INTERNAL_TOOLING_METRICS_HOSTNAME = "mongodb+srv://dev-metrics-pl-0.kewhj.mongodb.net"
INTERNAL_TOOLING_METRICS_USERNAME = "internal_tooling_user"
INTERNAL_TOOLING_METRICS_PASSWORD = "internal_tooling_user"


def get_mongo_metrics_client() -> MongoMetricsClient:
    """Construct client needed to setup metrics collection."""
    if not should_collect_internal_metrics():
        raise ExternalHostException("Metrics collection is only for MongoDB Virtual Workstations.")

    return MongoMetricsClient(
        mongo_client=pymongo.MongoClient(
            host=INTERNAL_TOOLING_METRICS_HOSTNAME,
            username=INTERNAL_TOOLING_METRICS_USERNAME,
            password=INTERNAL_TOOLING_METRICS_PASSWORD,
            socketTimeoutMS=1000,
            serverSelectionTimeoutMS=1000,
            connectTimeoutMS=1000,
            waitQueueTimeoutMS=1000,
            retryWrites=False,
            connect=False,
        ),
        db_name='metrics',
        collection_name='tooling_metrics',
        verbosity=MetricsVerbosity.LOG,
    )



from abc import abstractmethod
from typing import Any, Callable
from pydantic import BaseModel, Field, validate_model

from mongo_tooling_metrics.errors import InvalidMetricsSetup


class BaseHook(BaseModel):
    """Base class for hooks to collect intra-run function metrics."""

    original_fn: Callable[..., Any] = Field(..., exclude=True)

    @abstractmethod
    def passthrough_logic(self, *args, **kwargs) -> None:
        """Execute logic before calling the original function."""
        raise InvalidMetricsSetup(
            "'passthrough_logic' is called before calling the original function and must be defined."
        )

    def __call__(self, *args, **kwargs):
        """Call 'passthrough_logic' with **kwargs then call the 'original_fn' with **kwargs."""
        self.passthrough_logic(*args, **kwargs)
        self.original_fn(*args, **kwargs)

    def is_malformed(self) -> bool:
        """Make sure the hook still matches the pydantic model -- typing is not enforced intrarun by default."""
        *_, validation_error = validate_model(self.__class__, self.__dict__)
        return True if validation_error else False



from typing import Type

import pkg_resources
from mongo_tooling_metrics.base_hook import BaseHook
from mongo_tooling_metrics.registry import _HookRegistry

__version__ = pkg_resources.get_distribution('mongo-tooling-metrics').version
HOOK_REGISTRY = _HookRegistry()


def register_hook(hook_instance: BaseHook) -> BaseHook:
    """Add the hook to the registry if it doesn't exist & return the singleton hook instance."""
    return HOOK_REGISTRY._safe_set_hook(hook_instance)


def get_hook(hook_class: Type[BaseHook]) -> BaseHook:
    """Get the hook if it exists -- else fail."""
    return HOOK_REGISTRY._safe_get_hook(hook_class)


__all__ = ["base_models", "client", "errors", "hooks"]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mongo_tooling_metrics', 'mongo_tooling_metrics.lib']

package_data = \
{'': ['*']}

install_requires = \
['distro>=1.5.0,<2.0.0',
 'gitpython>=3.1.29,<4.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pymongo>=4.3.3,<5.0.0',
 'setuptools>=58.1.0,<59.0.0']

setup_kwargs = {
    'name': 'mongo-tooling-metrics',
    'version': '1.0.7',
    'description': 'A slim library which leverages Pydantic to reliably collect type enforced metrics and store them to MongoDB.',
    'long_description': 'None',
    'author': 'Tausif Rahman',
    'author_email': 'tausif.rahman@mongodb.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Common hooks that can be reused."""

from inspect import getcallargs
from mongo_tooling_metrics.base_hook import BaseHook


class ExitHook(BaseHook):
    """Plumb all sys.exit through this object so that we can access the exit code in atexit."""
    exit_code: int = 0

    def passthrough_logic(self, *args, **kwargs) -> None:
        """Store the exit code when sys.exit is called."""
        args_as_kwargs = getcallargs(self.original_fn, *args, **kwargs)
        self.exit_code = args_as_kwargs.get('status') if args_as_kwargs.get('status') != None else 0
